﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using UnityEditor;
using UnityEngine;
using Object = UnityEngine.Object;

#if UNITY_LOCALIZATION
using UnityEngine.Localization;
using UnityEngine.Localization.Components;
using UnityEngine.Localization.Settings;
#endif

namespace Magify
{
    public static class CreativeBuilder
    {
        private const string AssetBundleRootDirectory = "Bundles";
        private static readonly StringBuilder _errorReport = new();

        public static void BuildPlatform(BuildTarget buildTarget, ISet<string> builtCreatives)
        {
            BuildPlatform(buildTarget, buildTarget.ToString(), builtCreatives);
        }

        [MenuItem(PackageInfo.RootMenuItem + "Build Creatives Only For Editor")]
        public static void BuildAllAssetBundlesEditor()
        {
#if UNITY_EDITOR_WIN
            BuildPlatform(BuildTarget.StandaloneWindows64, "Editor", null);
#elif UNITY_EDITOR_OSX
            BuildPlatform(BuildTarget.StandaloneOSX, "Editor", null);
#elif UNITY_EDITOR_LINUX
            BuildPlatform(BuildTarget.StandaloneLinux64, "Editor", null);
#endif
        }

        public static bool HasError()
        {
            return _errorReport.Length > 0;
        }

        public static string GetErrorReport()
        {
            return _errorReport.ToString();
        }

        public static void BuildCreatives(BuildTarget buildTarget, string rootFolder, List<GameObject> creatives, ISet<string> builtCreatives)
        {
            _errorReport.Clear();

            var outputPath = Path.Combine(AssetBundleRootDirectory, rootFolder);
            if (Directory.Exists(outputPath))
            {
                Directory.Delete(outputPath, true);
            }
            Directory.CreateDirectory(outputPath);

            foreach (var creative in creatives)
            {
                var folderName = creative.name.ToLower();
                var path = Path.Combine(outputPath, folderName);
                Directory.CreateDirectory(path);

#if UNITY_LOCALIZATION
                var bundleLocalization = creative.GetComponent<BundleLocalization>();
                if (bundleLocalization != null)
                {
                    var availableLocales = bundleLocalization.GetAvailableLocales();
                    buildBundleForEachLocal(creative, path, availableLocales);
                }
                else
#endif
                {
                    buildBundle(creative, path, "en");
                }
            }

            CleanupBundlesFolder(outputPath);

#if UNITY_LOCALIZATION
            void buildBundleForEachLocal(GameObject creative, string path, List<Locale> availableLocales)
            {
                foreach (var locale in availableLocales)
                {
                    LocalizationSettings.SelectedLocale = locale;
                    buildBundle(creative, path, locale.Identifier.Code);
                }
            }
#endif

            void buildBundle(GameObject creative, string path, string cultureCode)
            {
                var assetPath = AssetDatabase.GetAssetPath(creative);
                try
                {
                    BuildCreative(creative, path, buildTarget, cultureCode);
                    builtCreatives?.Add(assetPath);
                }
                catch (Exception e)
                {
                    _errorReport.Append(e);
                    _errorReport.AppendLine();
                }
            }
        }

        public static void BuildCreative(GameObject creativePrefab, string outputPath, BuildTarget buildTarget, string cultureCode)
        {
            var tempPath = CopyCreative(creativePrefab);
            var bundles = new List<AssetBundleBuild>();
            var bundle = new AssetBundleBuild
            {
                assetBundleName = Guid.NewGuid() + ".bundle",
                assetNames = new[] { tempPath }
            };

            try
            {
#if UNITY_LOCALIZATION
                // We need to open and disable all localization, because even after removing it, it can block text changes in TextMeshPro
                // We do this in a separate EditPrefabContentsScope, because we need to open and close the prefab
                // so that the localization is no longer affected
                using (var editingScope = new PrefabUtility.EditPrefabContentsScope(tempPath))
                {
                    var bundleLocalization = editingScope.prefabContentsRoot.GetComponent<BundleLocalization>();
                    if (bundleLocalization != null)
                    {
                        Object.DestroyImmediate(bundleLocalization, true);
                    }

                    DisableAllLocalizeScripts(editingScope.prefabContentsRoot);
                }
#endif

                using (var editingScope = new PrefabUtility.EditPrefabContentsScope(tempPath))
                {
                    BundleFontCustomizer.PrepareFonts(editingScope.prefabContentsRoot);
                    RunAllCreativeProcessor(editingScope.prefabContentsRoot);
                    BundleFontCustomizer.RemoveLocalizationScripts(editingScope.prefabContentsRoot);
                }

                bundles.Add(bundle);
                BuildPipeline.BuildAssetBundles(outputPath, bundles.ToArray(), BuildAssetBundleOptions.AssetBundleStripUnityVersion, buildTarget);
                var oldPath = Path.Combine(outputPath, bundle.assetBundleName);
                var newPath = Path.Combine(outputPath, $"{cultureCode}.bundle");
                File.Move(oldPath, newPath);
            }
            finally
            {
                BundleFontCustomizer.RevertFontsSettings();
                AssetDatabase.DeleteAsset(tempPath);
                AssetDatabase.SaveAssets();
            }
        }

        public static void CreateLinkFile(ISet<string> assets)
        {
            var generator = new LinkXmlGenerator();
            generator.AddAssets(assets);
            generator.AddAssets(assets.SelectMany(AssetDatabase.GetDependencies));
            generator.Save(Path.Combine(AssetBundleRootDirectory, "link.xml"));
        }

        public static string CheckCreatives(List<GameObject> creatives)
        {
#if UNITY_LOCALIZATION
            LocalizationSettings.InitializationOperation.WaitForCompletion();
#endif

            var error = new StringBuilder();
            foreach (var creative in creatives)
            {
                var tempPath = string.Empty;
                try
                {
                    tempPath = CopyCreative(creative);
                    using var editingScope = new PrefabUtility.EditPrefabContentsScope(tempPath);
                    CheckFonts(editingScope.prefabContentsRoot);
                }
                catch (Exception e)
                {
                    error.Append(e.Message);
                    error.AppendLine();
                }
                finally
                {
                    BundleFontCustomizer.RevertFontsSettings();
                    if (tempPath != string.Empty)
                    {
                        AssetDatabase.DeleteAsset(tempPath);
                        AssetDatabase.SaveAssets();
                    }
                }
            }

            return error.ToString();
        }

#if UNITY_LOCALIZATION
        private static void DisableAllLocalizeScripts(GameObject rootPrefab)
        {
            foreach (var localizeStringEvent in rootPrefab.GetComponentsInChildren<LocalizeStringEvent>())
            {
                localizeStringEvent.enabled = false;
            }
        }
#endif

        private static void BuildPlatform(BuildTarget buildTarget, string rootFolder, ISet<string> builtCreatives)
        {
            var creatives = CollectCreatives();
            var duplicates = creatives.GroupBy(x => x.name.ToLower())
                .Where(g => g.Count() > 1)
                .Select(y => y.Key)
                .ToList();

            foreach (var name in duplicates)
            {
                Debug.LogError($"The name {name} used for more than one creative:");
                foreach (var creative in creatives.Where(c => c.name == name))
                {
                    Debug.LogError(AssetDatabase.GetAssetPath(creative), creative);
                }
            }

            if (duplicates.Count > 0)
            {
                return;
            }

#if UNITY_LOCALIZATION
            LocalizationSettings.InitializationOperation.WaitForCompletion();
#endif
            BuildCreatives(buildTarget, rootFolder, creatives, builtCreatives);

            if (HasError())
            {
                Debug.LogError(GetErrorReport());
            }
        }

        internal static List<GameObject> CollectCreatives()
        {
            Caching.ClearCache();

            var creatives = new List<GameObject>();
            foreach (var assetGuid in AssetDatabase.FindAssets("t:prefab"))
            {
                var assetPath = AssetDatabase.GUIDToAssetPath(assetGuid);
                var prefab = AssetDatabase.LoadAssetAtPath<GameObject>(assetPath);
                var prefabAssetType = PrefabUtility.GetPrefabAssetType(prefab);
                if (!(prefabAssetType == PrefabAssetType.Regular || prefabAssetType == PrefabAssetType.Variant))
                {
                    continue;
                }

                var oldCreative = prefab.GetComponent<BundledCreative>();
                var creative = prefab.GetComponent<BundledPrefab>();
                if (oldCreative != null || creative != null)
                {
                    creatives.Add(prefab);
                }
            }

            return creatives;
        }

        private static void RunAllCreativeProcessor(GameObject creative)
        {
            var buildCreativeProcess = creative.GetComponents<CreativeBuildProcessor>();
            foreach (var creativeProcessor in buildCreativeProcess)
            {
                creativeProcessor.ProcessCreative();
                Object.DestroyImmediate(creativeProcessor, true);
            }
        }

        private static void CleanupBundlesFolder(string path)
        {
            foreach (var file in Directory.GetFiles(path, "*", SearchOption.AllDirectories))
            {
                if (Path.GetFileNameWithoutExtension(file) == Path.GetFileName(Path.GetDirectoryName(file)) ||
                    Path.GetExtension(file).ToLower() == ".manifest")
                {
                    File.Delete(file);
                }
            }
        }

        private static void CheckFonts(GameObject creative)
        {
#if UNITY_LOCALIZATION
            var bundleLocalization = creative.GetComponent<BundleLocalization>();
            if (bundleLocalization != null)
            {
                var availableLocales = bundleLocalization.GetAvailableLocales();

                foreach (var locale in availableLocales)
                {
                    LocalizationSettings.SelectedLocale = locale;
                    BundleFontCustomizer.PrepareFonts(creative);
                    BundleFontCustomizer.RevertFontsSettings();
                }
            }
            else
#endif
            {
                BundleFontCustomizer.PrepareFonts(creative);
                BundleFontCustomizer.RevertFontsSettings();
            }
        }

        private static string GetTempPathForCreative(string originalPath)
        {
            return Path.Combine("Assets", Path.GetFileName(originalPath));
        }

        private static string CopyCreative(GameObject creative)
        {
            var originalPath = AssetDatabase.GetAssetPath(creative);
            var tempPath = GetTempPathForCreative(originalPath);
            var isCopied = AssetDatabase.CopyAsset(originalPath, tempPath);
            if (!isCopied)
            {
                throw new Exception($"Can't copy creative from {originalPath} to {tempPath}");
            }
            return tempPath;
        }

        private static void DeleteCreative(GameObject tempCreative)
        {
            var originalPath = AssetDatabase.GetAssetPath(tempCreative);
            var tempPath = GetTempPathForCreative(originalPath);
            AssetDatabase.DeleteAsset(tempPath);
            AssetDatabase.SaveAssets();
        }
    }
}