using System;
using System.Collections.Generic;
using System.Linq;
using JetBrains.Annotations;
using Magify.Model;
using UniRx;
using UnityEngine;

namespace Magify
{
    internal class PlatformDefault : PlatformAPI
    {
        [CanBeNull]
        private readonly EditorDevice _device;

        private readonly IDisposable _disposable;
        private readonly Subject<object> _onBackground = new();
        private readonly Subject<object> _onForeground = new();
        private readonly Subject<LocaleData> _onLocaleChanged = new();
        private readonly Subject<string> _onNewPackageInstalled = new();

        private readonly HashSet<string> _installedAppIdntifiers = new();
        private LocaleData _locale = new();

        public override RuntimePlatform RuntimePlatform => _device?.Platform ?? Application.platform;
        public override IObservable<object> OnBackground => _onBackground;
        public override IObservable<object> OnForeground => _onForeground;
        public override IObservable<LocaleData> OnLocaleChanged => _onLocaleChanged;
        public override IObservable<string> OnNewPackageInstalled => _onNewPackageInstalled;

        public override string DeviceName => _device?.DeviceName ?? "phone-xhdpi";
        public override string DeviceVersion => _device?.DeviceVersion ?? "30";
        public override bool IsTablet => _device?.IsTablet ?? base.IsTablet;
        public override Orientation SupportedOrientation => _device?.SupportedOrientation ?? base.SupportedOrientation;

        public PlatformDefault(EditorDevice device = null)
        {
            _device = device;
            CustomGpsAdId = _device?.GpsAdId;
            CustomIdfa = _device?.Idfa;
            CustomIdfv = _device?.Idfv;

            if (Application.isEditor && !Application.isPlaying)
            {
                return;
            }
            _disposable = Observable.EveryApplicationFocus().Subscribe(focus =>
            {
                if (!focus)
                {
                    _onBackground.OnNext(this);
                }
                else
                {
                    _onForeground.OnNext(this);
                }
            });
        }

        public override void AddPlatformSpecificData(AuthTokenPayload authToken, GeneralPrefs prefs, bool isGeoIpEnabled)
        {
            base.AddPlatformSpecificData(authToken, prefs, isGeoIpEnabled);
            if (_device?.Platform is RuntimePlatform.Android or RuntimePlatform.IPhonePlayer)
            {
                authToken.IsGeoIPEnabled = isGeoIpEnabled ? null : false;
            }
            if (_device?.Platform is RuntimePlatform.Android)
            {
                authToken.GpsAdId = prefs.GpsAdId.Value;
            }
            if (_device?.Platform is RuntimePlatform.IPhonePlayer)
            {
                authToken.IsIdfaAvailable = prefs.AttAuthorizationStatus.Value;
                authToken.Idfa = GetIdfa();
                authToken.Idfv = GetIdfv();
            }
        }

        public void GoToBackground()
        {
            _onBackground.OnNext(this);
        }

        public void ReturnToForeground()
        {
            _onForeground.OnNext(this);
        }

        public override LocaleData GetCurrentLocale()
        {
            return _locale;
        }

        public override void Dispose()
        {
            _disposable?.Dispose();
            base.Dispose();
        }

        public override void TweakUserLocale(string languageTag)
        {
            _locale = new LocaleData("US", languageTag);
            _onLocaleChanged.OnNext(_locale);
        }

        public override void TrackPromotedApplication(string bundleId)
        {
            if (bundleId == null) return;
            _installedAppIdntifiers.Add(bundleId);
            _onNewPackageInstalled.OnNext(bundleId);
        }

        public override bool IsApplicationInstalled(string identifier)
        {
            return _installedAppIdntifiers.Contains(identifier);
        }

        public override bool IsApplicationInstalled(PromotedApplication promotedApplication)
        {
            return _device?.Platform switch
            {
                RuntimePlatform.IPhonePlayer => promotedApplication?.Schemas?.Any(IsApplicationInstalled) ?? false,
                RuntimePlatform.Android or _ => IsApplicationInstalled(promotedApplication?.BundleId),
            };
        }
    }
}