#if UNITY_EDITOR || UNITY_IOS
using System;
using System.Linq;
using System.Collections.Generic;
using Magify.Model;
using UniRx;
using System.IO;
using JetBrains.Annotations;
using Newtonsoft.Json;
using UnityEngine;

namespace Magify
{
    internal partial class PlatformIOS : PlatformAPI, IInitializable
    {
        [NotNull]
        private static readonly MagifyLogger _migrationLogger = MagifyLogger.Get(LoggingScope.Migration);
        private static PlatformIOS _callbacksSingleton;
        private readonly IOSLegacyCountersStorage _legacyCountersStorage;
        private readonly Subject<object> _onBackground = new();
        private readonly Subject<object> _onForeground = new();
        private readonly Subject<object> _onApplicationDidBecomeActive = new();
        private readonly Subject<object> _onApplicationWillResignActive = new();
        private readonly Subject<LocaleData> _onLocaleChanged = new();
        private readonly Subject<string> _onNewPackageInstalled = new();
        [CanBeNull]
        private LocaleData _mockedLocale = null;

        public override RuntimePlatform RuntimePlatform => RuntimePlatform.IPhonePlayer;
        /// <remarks>
        /// Important: switching to main thread may cause foreground/background events to be skipped when enter background/overlapping with native UI.
        /// </remarks>
        public override IObservable<object> OnBackground => _onBackground;
        public override IObservable<object> OnForeground => _onForeground;
        public override IObservable<object> OnApplicationDidBecomeActive => _onApplicationDidBecomeActive;
        public override IObservable<object> OnApplicationWillResignActive => _onApplicationWillResignActive;
        public override IObservable<LocaleData> OnLocaleChanged => _onLocaleChanged.ObserveOnMainThread();
        public override IObservable<string> OnNewPackageInstalled => _onNewPackageInstalled.ObserveOnMainThread();
        public override string DeviceName => _magifyGetDeviceModel();
        public override string DeviceVersion => _magifyGetOsVersion();
        public override string BuildNumber => _magifyGetBuildNumber();

        public override bool IsTablet
        {
            get
            {
                var device = DeviceName;
                if (device.Contains("iPad"))
                {
                    return true;
                }
                if (device.Contains("iPhone"))
                {
                    return false;
                }
                return base.IsTablet;
            }
        }

        public PlatformIOS()
        {
            _callbacksSingleton = this;

            var countersBinaryStoragePath = Path.Combine(PackageInfo.PersistentPath, "iOS_legacy", "counters.bin");
            _legacyCountersStorage = new IOSLegacyCountersStorage(countersBinaryStoragePath);
            _magifyInitialize();
        }

        void IInitializable.Initialize()
        {
            _magifySetOnApplicationEnterBackground(OnApplicationEnterBackgroundHandler);
            _magifySetOnApplicationEnterForeground(OnApplicationEnterForegroundHandler);
            _magifySetOnApplicationDidBecomeActive(OnApplicationDidBecomeActiveHandler);
            _magifySetOnApplicationWillResignActive(OnApplicationWillResignActiveHandler);
        }

        public override LocaleData GetCurrentLocale()
        {
            return _mockedLocale ?? new LocaleData(_magifyGetLanguageWithRegion());
        }

        public override void AddPlatformSpecificData(AuthTokenPayload authToken, GeneralPrefs prefs, bool isGeoIpEnabled)
        {
            base.AddPlatformSpecificData(authToken, prefs, isGeoIpEnabled);
            // Pass null if isGeoIPEnabled == true (absent parameter will be treated by server as true)
            // In order to hide this parameter for sake of anti-fraud
            authToken.IsGeoIPEnabled = isGeoIpEnabled ? null : false;
            authToken.IsIdfaAvailable = prefs.AttAuthorizationStatus.Value;
            authToken.Idfa = GetIdfa();
            authToken.Idfv = GetIdfv();
        }

        public override void TweakUserLocale(string languageTag)
        {
            _mockedLocale = new LocaleData("US", languageTag);
            _onLocaleChanged?.OnNext(_mockedLocale);
        }

        public override void TrackPromotedApplication(string bundleId)
        {
            // Do nothing
        }

        public override bool IsApplicationInstalled(string urlScheme)
        {
            return _magifyIsApplicationInstalled(urlScheme);
        }

        public override bool IsApplicationInstalled(PromotedApplication application)
        {
            return application?.Schemas?.Any(_magifyIsApplicationInstalled) ?? false;
        }

        public override string GetIdfa()
        {
            return CustomIdfa ?? _magifyGetIdfa();
        }

        public override string GetIdfv()
        {
            return CustomIdfv ?? _magifyGetIdfv();
        }

        public override string GetEncodedDeviceIdentifier()
        {
            return _magifyGetEncodedDeviceIdentifier();
        }

        public override string GetValueInNativeStorage(string key)
        {
            try
            {
                return _magifyGetStringValueInKeychain(key);
            }
            catch (Exception e)
            {
                throw new MagifyFailedToAccessToNativeStorageException(
                    MagifyFailedToAccessToNativeStorageException.NativeStorageKind.IOSKeychain,
                    MagifyFailedToAccessToNativeStorageException.NativeStorageAccessKind.Get,
                    key,
                    e);
            }
        }

        public override void SetValueInNativeStorage(string value, string key)
        {
            try
            {
                _magifySetStringValueInKeychain(value, key);
            }
            catch (Exception e)
            {
                throw new MagifyFailedToAccessToNativeStorageException(
                    MagifyFailedToAccessToNativeStorageException.NativeStorageKind.IOSKeychain,
                    MagifyFailedToAccessToNativeStorageException.NativeStorageAccessKind.Set,
                    key,
                    e);
            }
        }

        public override void RemoveValueInNativeStorage(string key)
        {
            try
            {
                _magifyRemoveValueInKeychain(key);
            }
            catch (Exception e)
            {
                throw new MagifyFailedToAccessToNativeStorageException(
                    MagifyFailedToAccessToNativeStorageException.NativeStorageKind.IOSKeychain,
                    MagifyFailedToAccessToNativeStorageException.NativeStorageAccessKind.Remove,
                    key,
                    e);
            }
        }

        public override void ClearNativeStorage()
        {
            _magifyClearKeychains();
        }

        [AOT.MonoPInvokeCallback(typeof(Action))]
        private static void OnApplicationEnterBackgroundHandler()
        {
            _callbacksSingleton._onBackground.OnNext(_callbacksSingleton);
        }

        [AOT.MonoPInvokeCallback(typeof(Action))]
        private static void OnApplicationEnterForegroundHandler()
        {
            _callbacksSingleton._onForeground.OnNext(_callbacksSingleton);
        }

        [AOT.MonoPInvokeCallback(typeof(Action))]
        private static void OnApplicationDidBecomeActiveHandler()
        {
            _callbacksSingleton._onApplicationDidBecomeActive.OnNext(_callbacksSingleton);
        }

        [AOT.MonoPInvokeCallback(typeof(Action))]
        private static void OnApplicationWillResignActiveHandler()
        {
            _callbacksSingleton._onApplicationWillResignActive.OnNext(_callbacksSingleton);
        }

        public override void Dispose()
        {
            _magifyDispose();
            _legacyCountersStorage.Dispose();
            _onBackground.Dispose();
            _onForeground.Dispose();
            _onApplicationWillResignActive.Dispose();
            _onLocaleChanged.Dispose();
            _onNewPackageInstalled.Dispose();
            base.Dispose();
        }

        public override MigrationData LoadLegacyNativeDataForMigration()
        {
            var legacyDirPath = _magifyCollectMigrationData();

            if (!Directory.Exists(legacyDirPath))
            {
                return null;
            }

            var countersFilePath = Path.Combine(legacyDirPath, "counters");
            if (File.Exists(countersFilePath))
            {
                MigrateLegacyCountersToBinaryStorageFromFile(countersFilePath);
            }

            var migrationData = new MigrationData();
            FillMigrationDataFromLegacyDirectory(migrationData, legacyDirPath);

            return migrationData;
        }

        public override void CleanLegacyNativeDataAfterMigration()
        {
            _magifyClearDataAfterMigration();
        }

        [CanBeNull]
        public override int? GetGlobalCounterFromLegacyData(CounterType type, CounterKey key)
        {
            _migrationLogger.Log($"Getting global counter for type: {type}, key: {key}");
            var (scope, keyStr) = type switch
            {
                CounterType.Bonuses => (default, default),
                CounterType.Clicks => ("GlobalClickCounter", ExtractClicksCounterKeyFrom(key)),
                CounterType.Events => ("GlobalEventCounter", ExtractEventsCounterKeyFrom(key)),
                CounterType.Impressions => key.LimitId == null
                    ? ("GlobalImpressionCounter", ExtractImpressionsCounterKeyFrom(key))
                    : ("PeriodTempLimitsCounter", ExtractTempLimitsCounterKeyFrom(key)),
                CounterType.Rewards => ("GlobalRVCounter", ExtractRewardsCounterKeyFrom(key)),
                CounterType.Nested => ("GlobalNestedCampaignsCounter", ExtractNestedCounterKeyFrom(key)),
                CounterType.LtoStart => ("LtoActivationPerTriggerGlobalCounter", ExtractLtoCounterKeyFrom(key)),
                _ => (default, default),
            };

            if (scope == null) _migrationLogger.Log($"The {nameof(scope)} to get global counter is null, returning null");
            if (keyStr == null) _migrationLogger.Log($"The {nameof(keyStr)} to get global counter is null, returning null");
            if (scope == null || keyStr == null)
            {
                return null;
            }

            return _legacyCountersStorage.GetGlobalCounter(scope, keyStr);
        }

        private void MigrateLegacyCountersToBinaryStorageFromFile(string countersFilePath)
        {
            var countersJson = File.ReadAllText(countersFilePath);
            var counters = JsonConvert.DeserializeObject<Dictionary<string, Dictionary<string, int>>>(countersJson);

            _legacyCountersStorage.MigrateCounters(counters);
        }

        private static void FillMigrationDataFromLegacyDirectory(MigrationData migrationData, string legacyDirectoryPath)
        {
            var contextFilePath = Path.Combine(legacyDirectoryPath, "context");
            if (File.Exists(contextFilePath))
            {
                migrationData.ContextPath = contextFilePath;
            }

            var ltoInfoFilePath = Path.Combine(legacyDirectoryPath, "ltoInfo");
            if (File.Exists(ltoInfoFilePath))
            {
                var ltoInfoJson = File.ReadAllText(ltoInfoFilePath);
                var ltoInfos = JsonConvert.DeserializeObject<List<LtoInfoMigrationData>>(ltoInfoJson);

                migrationData.ActiveTimeLimitedOffers = ltoInfos.Select(offer => new LtoModel
                {
                    HasPriority = offer.IsPriorityLaunch,
                    Trigger = offer.LaunchedTrigger,
                    BadgePlaceholder = offer.DefaultSpotImage,
                    CampaignName = offer.CampaignName,
                    Spot = offer.Spot,
                    StartTime = DateTime.UnixEpoch + TimeSpan.FromSeconds(offer.StartTime),
                    EndTime = DateTime.UnixEpoch + TimeSpan.FromSeconds(offer.EndTime)
                }).ToList();
            }

            var cachedAnalyticsFilePath = Path.Combine(legacyDirectoryPath, "cachedAnalytics");
            if (File.Exists(cachedAnalyticsFilePath))
            {
                var cachedAnalyticsJson = File.ReadAllText(cachedAnalyticsFilePath);
                var cachedAnalytics = JsonConvert.DeserializeObject<CachedEventsMigrationData>(cachedAnalyticsJson);

                migrationData.AppLaunchEvents = cachedAnalytics.AppLaunchEvents;
                migrationData.CustomSessionEvents = cachedAnalytics.SessionEventEvents;
                migrationData.AppBackgroundEvents = cachedAnalytics.AppBackgroundEvents;
                migrationData.TrackInstallEvents = cachedAnalytics.TrackInstallEvents;
                migrationData.TransactionEvents = cachedAnalytics.TransactionEvents;
                migrationData.FirebaseMappingEvents = cachedAnalytics.FirebaseMappingEvents;
                migrationData.MailingStatusMappingEvents = cachedAnalytics.MailingStatusMappingEvents;
                migrationData.ImpressionEvents = cachedAnalytics.ImpressionEvents;
                migrationData.ApplovinAdsImpressionEvents = cachedAnalytics.ApplovinAdsImpressionEvents;
                migrationData.IronSourceAdsImpressionEvents = cachedAnalytics.IronSourceAdsImpressionEvents;
                migrationData.ImpressionFailEvents = cachedAnalytics.ImpressionFailEvents;
                migrationData.TrialActivationEvents = cachedAnalytics.TrialActivationEvents;
                migrationData.PaidSubscriptionActivationEvents = cachedAnalytics.PaidSubscriptionActivationEvents;
                migrationData.InAppEvents = cachedAnalytics.InAppEvents;
                migrationData.ClickEvents = cachedAnalytics.ClickEvents;
                migrationData.AdsClickEvents = cachedAnalytics.AdsClickEvents;
                migrationData.PurchasedRecords = cachedAnalytics.PurchasedRecords;
            }

            var generalPrefsFilePath = Path.Combine(legacyDirectoryPath, "generalInfo");
            if (File.Exists(generalPrefsFilePath))
            {
                var generalPrefsJson = File.ReadAllText(generalPrefsFilePath);
                _migrationLogger.Log($"Native GeneralPrefs json content: \n{generalPrefsJson}");
                var generalPrefs = JsonConvert.DeserializeObject<GeneralPrefsMigrationData>(generalPrefsJson);

                var purchasedInAppProducts = generalPrefs.PurchasedInAppProducts;
                var inAppStatus = InAppStatus.NeverPurchased;
                if (purchasedInAppProducts is { Count: > 0 })
                {
                    inAppStatus = InAppStatus.Purchased;
                }

                migrationData.ClientId = generalPrefs.ClientId;
                migrationData.FirstInstalledVersion = generalPrefs.FirstInstalledVersion;
                migrationData.FirstLaunchDate = DateTime.UnixEpoch + TimeSpan.FromSeconds(generalPrefs.FirstLaunchDate);
                migrationData.InAppStatus = inAppStatus;
                migrationData.AdjustId = generalPrefs.AdjustId;
                migrationData.AttAuthorizationStatus = generalPrefs.AttAuthorizationStatus;
                migrationData.UsedDefaultCampaignTypes = generalPrefs.UsedDefaultCampaignTypes;
                migrationData.ProductsReceiptTimestamps = generalPrefs.ProductsTimestamps?.ToDictionary(
                    pair => pair.Key,
                    pair => DateTime.UnixEpoch + TimeSpan.FromSeconds(pair.Value));
                migrationData.PurchasedInAppProducts = purchasedInAppProducts;
                migrationData.PurchasedSubscriptionProducts = generalPrefs.PurchasedSubscriptionProducts;
                migrationData.UsedFreeProducts = generalPrefs.UsedFreeProducts;
                migrationData.TrackedAppInstalls = generalPrefs.TrackedAppInstalls;
                migrationData.FirebaseInstanceId = generalPrefs.FirebaseInstanceId;
                migrationData.GlobalSessionCounter = generalPrefs.GlobalSessionCounter;
                migrationData.SessionId = generalPrefs.SessionId;
                migrationData.SessionStarted = generalPrefs.SessionStarted;
                migrationData.LastActivity = generalPrefs.LastActivity;
                migrationData.MediaSource = new MediaSource(
                    generalPrefs.MediaSourceMigrationData?.NetworkName,
                    generalPrefs.MediaSourceMigrationData?.CampaignName,
                    generalPrefs.MediaSourceMigrationData?.GroupName
                );
            }
        }

        private static string ExtractClicksCounterKeyFrom(CounterKey key)
        {
            switch (key.GetUsedProperties())
            {
                case CounterKey.Flags.CampaignNameProp | CounterKey.Flags.SourceProp | CounterKey.Flags.SourceTypeProp:
                    return $"{key.SourceType!.Value.ToEnumString()}-{key.CampaignName}-{key.Source}";

                case CounterKey.Flags.CampaignNameProp:
                    return key.CampaignName;
            }

            _migrationLogger.Log($"{nameof(ExtractClicksCounterKeyFrom)} called for key: {key}, flags: {key.GetUsedProperties()}");
            return null;
        }

        private static string ExtractEventsCounterKeyFrom(CounterKey key)
        {
            switch (key.GetUsedProperties())
            {
                case CounterKey.Flags.SourceProp:
                    return key.Source;
            }

            _migrationLogger.Log($"{nameof(ExtractEventsCounterKeyFrom)} called for key: {key}, flags: {key.GetUsedProperties()}");
            return null;
        }

        private static string ExtractImpressionsCounterKeyFrom(CounterKey key)
        {
            switch (key.GetUsedProperties())
            {
                case CounterKey.Flags.CampaignNameProp | CounterKey.Flags.SourceProp | CounterKey.Flags.SourceTypeProp:
                    switch (key.SourceType)
                    {
                        case SourceType.Event:
                            return $"{key.CampaignName}-{key.Source}";

                        case SourceType.Trigger:
                            return $"{key.CampaignName}-trigger-{key.Source}";

                        case SourceType.Placement:
                            return null;
                    }
                    break;

                case CounterKey.Flags.CampaignNameProp | CounterKey.Flags.SourceTypeProp:
                    return $"{key.CampaignName}-{key.SourceType!.Value.ToEnumString()}";

                case CounterKey.Flags.CampaignTypeProp | CounterKey.Flags.SourceTypeProp:
                    return $"{key.CampaignType!.Value.ToEnumString()}-{key.SourceType!.Value.ToEnumString()}";

                case CounterKey.Flags.CampaignTypeProp:
                    return key.CampaignType!.Value.ToEnumString();

                case CounterKey.Flags.CampaignNameProp:
                    return key.CampaignName;

                case CounterKey.Flags.SourceTypeProp:
                    return key.SourceType!.Value.ToEnumString();
            }

            _migrationLogger.Log($"{nameof(ExtractImpressionsCounterKeyFrom)} called for key: {key}, flags: {key.GetUsedProperties()}");
            return null;
        }

        private static string ExtractTempLimitsCounterKeyFrom(CounterKey key)
        {
            switch (key.GetUsedProperties())
            {
                case CounterKey.Flags.LimitIdProp | CounterKey.Flags.CampaignTypeProp:
                    return $"tempLimits-{key.LimitId}-{key.CampaignType!.Value.ToEnumString()}";
            }

            _migrationLogger.Log($"{nameof(ExtractTempLimitsCounterKeyFrom)} called for key: {key}, flags: {key.GetUsedProperties()}");
            return null;
        }

        private static string ExtractRewardsCounterKeyFrom(CounterKey key)
        {
            switch (key.GetUsedProperties())
            {
                case CounterKey.Flags.CampaignNameProp | CounterKey.Flags.SourceProp | CounterKey.Flags.SourceTypeProp:
                    return $"rv-{key.SourceType!.Value.ToEnumString()}-{key.CampaignName}-{key.Source}";

                case CounterKey.Flags.CampaignNameProp:
                    return key.CampaignName;
            }

            _migrationLogger.Log($"{nameof(ExtractRewardsCounterKeyFrom)} called for key: {key}, flags: {key.GetUsedProperties()}");
            return null;
        }

        private static string ExtractNestedCounterKeyFrom(CounterKey key)
        {
            switch (key.GetUsedProperties())
            {
                case CounterKey.Flags.NestedNameProp:
                    return key.NestedName;
            }

            _migrationLogger.Log($"{nameof(ExtractNestedCounterKeyFrom)} called for key: {key}, flags: {key.GetUsedProperties()}");
            return null;
        }

        private static string ExtractLtoCounterKeyFrom(CounterKey key)
        {
            switch (key.GetUsedProperties())
            {
                case CounterKey.Flags.CampaignNameProp | CounterKey.Flags.SourceProp:
                    return $"{key.CampaignName}-{key.Source}";

                case CounterKey.Flags.CampaignNameProp:
                    return $"lto_actiovation_counter_{key.CampaignName}";
            }

            _migrationLogger.Log($"{nameof(ExtractLtoCounterKeyFrom)} called for key: {key}, flags: {key.GetUsedProperties()}");
            return null;
        }
    }
}
#endif