﻿using System;
using System.Threading;
using JetBrains.Annotations;
using UniRx;

namespace Magify
{
    internal class SyncTimer : ICancelable
    {
        private readonly int _minSyncInterval;
        private System.Timers.Timer _timer = new();
        private Action _onElapsedAction;

        public bool IsDisposed => _onElapsedAction == null;

        public SyncTimer([NotNull] Action onElapsedAction, int minSyncInterval = 1000)
        {
            if (minSyncInterval <= 0)
            {
                throw new ArgumentOutOfRangeException($"{nameof(minSyncInterval)} cannot be equal to or less than 0");
            }
            _onElapsedAction = onElapsedAction;
            _minSyncInterval = minSyncInterval;
        }

        public void Start(int syncInterval)
        {
            ThrowIfDisposed();

            _timer.Interval = Math.Max(_minSyncInterval, syncInterval);
            _timer.Elapsed += async (_, _) =>
            {
                await TaskScheduler.SwitchToMainThread(CancellationToken.None);
                _onElapsedAction();
                _timer.Start();
            };
            _timer.Start();
        }

        public void Restart(int? syncInterval = null)
        {
            ThrowIfDisposed();

            _timer.Stop();
            if (syncInterval.HasValue)
            {
                _timer.Interval = Math.Max(_minSyncInterval, syncInterval.Value);
            }
            _timer.Start();
        }

        public void Stop()
        {
            ThrowIfDisposed();
            _timer.Stop();
        }

        public void Dispose()
        {
            if (IsDisposed) return;

            _timer.Stop();
            _onElapsedAction = null;
            _timer = null;
        }

        private void ThrowIfDisposed()
        {
            if (IsDisposed)
            {
                throw new MagifyObjectDisposedException(nameof(SyncTimer));
            }
        }
    }
}