using System;
using Cysharp.Threading.Tasks;
using UniRx;
using UnityEngine;
using static Magify.MagifyManager;

namespace Magify
{
    public class NetworkStatus : INetworkStatusProvider, IDisposable
    {
        private static readonly MagifyLogger _logger = MagifyLogger.Get(MagifyService.LogScope);

        private const float MinUpdateInterval = 1;
        private const float MaxUpdateInterval = 5;

        private float _lastUpdate;
        private bool _disposed;

        public IReactiveProperty<NetworkState> Reachability { get; } = new ReactiveProperty<NetworkState>(NetworkState.NotReachable);

        internal NetworkStatus()
        {
            _lastUpdate = -MinUpdateInterval;
            CheckConnectionRoutine().Forget();
            Reachability
                .TakeWhile(_ => !_disposed)
                .Subscribe(_ => { _logger.Log($"{nameof(NetworkState)} changed to {Reachability.Value}"); });
        }

        private void UpdateNetwork()
        {
            if (Time.realtimeSinceStartup - _lastUpdate <= MinUpdateInterval) return;
            var reachable = Application.internetReachability != NetworkReachability.NotReachable;
            switch (Reachability.Value)
            {
                case NetworkState.Reachable when !reachable:
                    Reachability.Value = NetworkState.NotReachable;
                    break;
                case NetworkState.NotReachable when reachable:
                    Reachability.Value = NetworkState.Reachable;
                    break;
            }

            _lastUpdate = Time.realtimeSinceStartup;
        }

        private async UniTask CheckConnectionRoutine()
        {
            while (!_disposed)
            {
                UpdateNetwork();
                while (_lastUpdate + MaxUpdateInterval > Time.realtimeSinceStartup)
                {
                    await UniTask.Yield();
                }
            }
        }

        public void Dispose()
        {
            _disposed = true;
        }
    }
}