﻿using System;
using System.Globalization;
using System.Threading;
using Cysharp.Threading.Tasks;
using Newtonsoft.Json;
using UnityEngine;
using UnityEngine.Networking;
using UnityEngine.Scripting;

namespace Magify
{
    internal class WorldTimeApiServerTimeProvider : IServerTimeProvider
    {
        private static readonly MagifyLogger _logger = MagifyLogger.Get(LoggingScope.Network);

        [Preserve]
        public struct TimeData
        {
            [JsonProperty("utc_datetime"), Preserve]
            public string UTCDateTime { get; set; }
        }

        public async UniTask<DateTime?> TryGetServerTime(CancellationToken cancellationToken)
        {
            const string url = "https://worldtimeapi.org/api/ip";
            var request = UnityWebRequest.Get(url);
            await request.SendWebRequest().WithCancellation(cancellationToken);

            if (request.result is not UnityWebRequest.Result.Success)
            {
                _logger.Log($"Failed to get server time from WorldTimeApi with result: {request.result}, and error: {request.error}");
                return null;
            }

            try
            {
                _logger.Log($"Response time from WorldTimeApi: {request.downloadHandler.text}");
                var timeData = JsonFacade.DeserializeObject<TimeData>(request.downloadHandler.text);
                var utcDateTime = timeData.UTCDateTime.Replace("+00:00", "Z");
                var parsed = DateTime.Parse(utcDateTime);
                _logger.Log($"Parsed time from WorldTimeApi: UTC: {parsed.ToUniversalTime()}, Local: {parsed.ToLocalTime()}");
                return parsed.ToUniversalTime();
            }
            catch (Exception e)
            {
                _logger.Log($"Exception during parsing server time from WorldTimeApi: {e.Message}");
                _logger.LogException(e);
            }
            return null;
        }
    }
}