using System;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;
using Magify.Types;
using Cysharp.Threading.Tasks;
using FluentAssertions;
using Magify.Model;
using NSubstitute;
using UniRx;
using NUnit.Framework;
using UnityEditor;
using UnityEngine.Networking;

namespace Magify.Tests
{
    public class AnalyticsTrackerTests
    {
        private PlatformAPI _platformAPI;
        private NetworkClient _realNetworkClient;
        private WebResponseMessage _interceptedResponseMessage;

        private CancellationTokenSource _timeoutCts;

        private static EndpointUrl StagingUrl => EndpointUrl.Staging;
        private IAuthorizationConfigBuilder AuthConfig => NetworkTestEnvironment.AuthConfig;

        [SetUp]
        public void Setup()
        {
            EditorModeTestsEnvironment.Clear();
            _platformAPI = new PlatformDefault();
            _realNetworkClient = new NetworkClient();
            _interceptedResponseMessage = default;
            _timeoutCts = new CancellationTokenSource(TimeSpan.FromSeconds(3));
        }

        [TearDown]
        public void Cleanup()
        {
            _platformAPI.Dispose();
            _timeoutCts.Cancel();
            _timeoutCts.Dispose();
            EditorModeTestsEnvironment.Clear();
        }

        [Test]
        [UnitTestTarget(typeof(AnalyticsTracker), nameof(AnalyticsTracker.TrackCustomSessionEvent))]
        public async Task TrackCustomSessionEvent_ThenShouldSucceedSendEventToServer()
        {
            // Arrange
            using var subsystems = CreateAnalyticsTracker(out var analyticsTracker, out var networkClient, out _);
            subsystems.ForEach<IInitializable>(s => s.Initialize());

            // Act
            analyticsTracker.TrackCustomSessionEvent("custom_session_event", new Dictionary<string, object>());
            await TaskHelper.WaitUntil(() => _interceptedResponseMessage.Result != UnityWebRequest.Result.InProgress, _timeoutCts.Token);

            // Assert
            FailIfTimeoutTokenWasCanceled();
            networkClient
                .Received(2)
                .SendAsync(Arg.Any<string>(), Arg.Any<WebRequestMessage>(), Arg.Any<CancellationToken>()).Forget();
            _interceptedResponseMessage.ResponseCode.Should().Be(200);
            _interceptedResponseMessage.Result.Should().Be(UnityWebRequest.Result.Success);
        }

        [Test]
        [UnitTestTarget(typeof(AnalyticsTracker), nameof(AnalyticsTracker.TrackAppInstallEvent))]
        public async Task TrackAppInstallEvent_ThenShouldSucceedSendEventToServer()
        {
            // Arrange
            using var subsystems = CreateAnalyticsTracker(out var analyticsTracker, out var networkClient, out _);
            subsystems.ForEach<IInitializable>(s => s.Initialize());

            // Act
            analyticsTracker.TrackAppInstallEvent(new HashSet<string> { "app" });
            await TaskHelper.WaitUntil(() => _interceptedResponseMessage.Result != UnityWebRequest.Result.InProgress, _timeoutCts.Token);

            // Assert
            FailIfTimeoutTokenWasCanceled();
            networkClient
                .Received(2)
                .SendAsync(Arg.Any<string>(), Arg.Any<WebRequestMessage>(), Arg.Any<CancellationToken>()).Forget();
            _interceptedResponseMessage.ResponseCode.Should().Be(200);
            _interceptedResponseMessage.Result.Should().Be(UnityWebRequest.Result.Success);
        }

        [Test]
        [UnitTestTarget(typeof(AnalyticsTracker), nameof(AnalyticsTracker.TrackTransaction))]
        public async Task TrackIncomeTransaction_ThenShouldSucceedSendEventToServer()
        {
            // Arrange
            using var subsystems = CreateAnalyticsTracker(out var analyticsTracker, out var networkClient, out _);
            subsystems.ForEach<IInitializable>(s => s.Initialize());

            // Act
            analyticsTracker.TrackTransaction(
                Transaction.CreateIncome("source"),
                new List<BonusInfo>
                {
                    new()
                    {
                        BonusName = "bonus"
                    }
                },
                new ProductInfo
                {
                    ProductId = "product",
                    Price = "1",
                    Currency = "USD"
                },
                new CampaignImpression
                {
                    CampaignName = "test",
                    EventName = "test",
                    EventNumber = 1,
                    Parameters = new Dictionary<string, object>()
                });
            await TaskHelper.WaitUntil(() => _interceptedResponseMessage.Result != UnityWebRequest.Result.InProgress, _timeoutCts.Token);

            // Assert
            FailIfTimeoutTokenWasCanceled();
            networkClient
                .Received(2)
                .SendAsync(Arg.Any<string>(), Arg.Any<WebRequestMessage>(), Arg.Any<CancellationToken>()).Forget();
            _interceptedResponseMessage.ResponseCode.Should().Be(200);
            _interceptedResponseMessage.Result.Should().Be(UnityWebRequest.Result.Success);
        }

        [Test]
        [UnitTestTarget(typeof(AnalyticsTracker), nameof(AnalyticsTracker.TrackFirebaseMappingEvent))]
        public async Task TrackFirebaseMappingEvent_ThenShouldSucceedSendEventToServer()
        {
            // Arrange
            using var subsystems = CreateAnalyticsTracker(out var analyticsTracker, out var networkClient, out _);
            subsystems.ForEach<IInitializable>(s => s.Initialize());

            // Act
            analyticsTracker.TrackFirebaseMappingEvent("123456");
            await TaskHelper.WaitUntil(() => _interceptedResponseMessage.Result != UnityWebRequest.Result.InProgress, _timeoutCts.Token);

            // Assert
            FailIfTimeoutTokenWasCanceled();
            networkClient
                .Received(2)
                .SendAsync(Arg.Any<string>(), Arg.Any<WebRequestMessage>(), Arg.Any<CancellationToken>()).Forget();
            _interceptedResponseMessage.ResponseCode.Should().Be(200);
            _interceptedResponseMessage.Result.Should().Be(UnityWebRequest.Result.Success);
        }

        [Test]
        [UnitTestTarget(typeof(AnalyticsTracker), nameof(AnalyticsTracker.TrackMailingStatusMappingEvent))]
        public async Task TrackMailingStatusMappingEvent_ThenShouldSucceedSendEventToServer()
        {
            // Arrange
            using var subsystems = CreateAnalyticsTracker(out var analyticsTracker, out var networkClient, out _);
            subsystems.ForEach<IInitializable>(s => s.Initialize());

            // Act
            analyticsTracker.TrackMailingStatusMappingEvent("test@gmail.com");
            await TaskHelper.WaitUntil(() => _interceptedResponseMessage.Result != UnityWebRequest.Result.InProgress, _timeoutCts.Token);

            // Assert
            FailIfTimeoutTokenWasCanceled();
            networkClient
                .Received(2)
                .SendAsync(Arg.Any<string>(), Arg.Any<WebRequestMessage>(), Arg.Any<CancellationToken>()).Forget();
            _interceptedResponseMessage.ResponseCode.Should().Be(200);
            _interceptedResponseMessage.Result.Should().Be(UnityWebRequest.Result.Success);
        }

        [Test]
        [UnitTestTarget(typeof(AnalyticsTracker), nameof(AnalyticsTracker.TrackApplovinAdsImpression))]
        public async Task TrackApplovinImpressionEvent_ThenShouldSucceedSendEventToServer()
        {
            // Arrange
            using var subsystems = CreateAnalyticsTracker(out var analyticsTracker, out var networkClient, out _);
            subsystems.ForEach<IInitializable>(s => s.Initialize());

            var impression = new CampaignImpression
            {
                CampaignName = "test",
                EventName = "test",
                EventNumber = 1,
                Parameters = new Dictionary<string, object>()
            };
            var applovinData = new ApplovinAdsImpression
            {
                Id = "testId",
                Network = "testNetwork",
                ImpressionId = GUID.Generate().ToString(),
                Revenue = 0.1,
                Precision = "test"
            };

            // Act
            analyticsTracker.TrackApplovinAdsImpression(impression, applovinData);
            await TaskHelper.WaitUntil(() => _interceptedResponseMessage.Result != UnityWebRequest.Result.InProgress, _timeoutCts.Token);

            // Assert
            FailIfTimeoutTokenWasCanceled();
            networkClient
                .Received(2)
                .SendAsync(Arg.Any<string>(), Arg.Any<WebRequestMessage>(), Arg.Any<CancellationToken>()).Forget();
            _interceptedResponseMessage.ResponseCode.Should().Be(200);
            _interceptedResponseMessage.Result.Should().Be(UnityWebRequest.Result.Success);
        }

        [Test]
        [UnitTestTarget(typeof(AnalyticsTracker), nameof(AnalyticsTracker.TrackIronSourceAdsImpression))]
        public async Task TrackIronSourceImpressionEvent_ThenShouldSucceedSendEventToServer()
        {
            // Arrange
            using var subsystems = CreateAnalyticsTracker(out var analyticsTracker, out var networkClient, out _);
            subsystems.ForEach<IInitializable>(s => s.Initialize());

            var impression = new CampaignImpression
            {
                CampaignName = "test",
                EventName = "test",
                EventNumber = 1
            };
            var applovinData = new IronSourceAdsImpression
            {
                Id = "testId",
                Network = "testNetwork",
                AdUnit = "banner",
                Revenue = 0.1,
                Precision = "test"
            };

            // Act
            analyticsTracker.TrackIronSourceAdsImpression(impression, applovinData);
            await TaskHelper.WaitUntil(() => _interceptedResponseMessage.Result != UnityWebRequest.Result.InProgress, _timeoutCts.Token);

            // Assert
            FailIfTimeoutTokenWasCanceled();
            networkClient
                .Received(2)
                .SendAsync(Arg.Any<string>(), Arg.Any<WebRequestMessage>(), Arg.Any<CancellationToken>()).Forget();
            _interceptedResponseMessage.ResponseCode.Should().Be(200);
            _interceptedResponseMessage.Result.Should().Be(UnityWebRequest.Result.Success);
        }

        [Test]
        [UnitTestTarget(typeof(AnalyticsTracker), nameof(AnalyticsTracker.TrackPurchaseEvent))]
        public async Task TrackSubscriptionPurchaseEvent_ThenShouldSucceedSendEventToServer()
        {
            // Arrange
            using var subsystems = CreateAnalyticsTracker(out var analyticsTracker, out var networkClient, out _);
            subsystems.ForEach<IInitializable>(s => s.Initialize());

            var impression = new CampaignImpression
            {
                CampaignName = "test",
                EventName = "test",
                EventNumber = 1
            };
            var product = new ProductInfo
            {
                ProductId = "testId",
                Price = "9.99",
                Currency = "USD"
            };
            var purchase = new PurchaseInfo
            {
                Product = product
            };

            // Act
            analyticsTracker.TrackPurchaseEvent(EventType.TrialActivation, impression, purchase);
            await TaskHelper.WaitUntil(() => _interceptedResponseMessage.Result != UnityWebRequest.Result.InProgress, _timeoutCts.Token);

            // Assert
            FailIfTimeoutTokenWasCanceled();
            networkClient
                .Received(2)
                .SendAsync(Arg.Any<string>(), Arg.Any<WebRequestMessage>(), Arg.Any<CancellationToken>()).Forget();
            _interceptedResponseMessage.ResponseCode.Should().Be(200);
            _interceptedResponseMessage.Result.Should().Be(UnityWebRequest.Result.Success);
        }

        [Test]
        [UnitTestTarget(typeof(AnalyticsTracker), nameof(AnalyticsTracker.TrackCommonEvent))]
        public async Task TrackCampaignImpression_ThenShouldSucceedSendEventToServer()
        {
            // Arrange
            using var subsystems = CreateAnalyticsTracker(out var analyticsTracker, out var networkClient, out _);
            subsystems.ForEach<IInitializable>(s => s.Initialize());

            var impression = new CampaignImpression
            {
                CampaignName = "testCampaign",
                EventName = "testEvent",
                Parameters = new Dictionary<string, object>(),
                ImpressionNumber = 1,
                SessionImpressionNumber = 1,
                CampaignTypeImpressionNumber = 1,
                CampaignTypeSessionImpressionNumber = 1,
                EventNumber = 1,
                SessionEventNumber = 1,
            };

            // Act
            analyticsTracker.TrackCommonEvent(EventType.Impression, impression);
            await TaskHelper.WaitUntil(() => _interceptedResponseMessage.Result != UnityWebRequest.Result.InProgress, _timeoutCts.Token);

            // Assert
            FailIfTimeoutTokenWasCanceled();
            networkClient
                .Received(2)
                .SendAsync(Arg.Any<string>(), Arg.Any<WebRequestMessage>(), Arg.Any<CancellationToken>()).Forget();
            _interceptedResponseMessage.ResponseCode.Should().Be(200);
            _interceptedResponseMessage.Result.Should().Be(UnityWebRequest.Result.Success);
        }

        [Test]
        [UnitTestTarget(typeof(AnalyticsTracker), nameof(AnalyticsTracker.TrackImpressionFailEvent))]
        public async Task TrackImpressionFailEvent_ThenShouldSucceedSendEventToServer()
        {
            // Arrange
            using var subsystems = CreateAnalyticsTracker(out var analyticsTracker, out var networkClient, out _);
            subsystems.ForEach<IInitializable>(s => s.Initialize());

            var impression = new CampaignImpression
            {
                CampaignName = "testCampaign",
                EventName = "testEvent",
                Parameters = new Dictionary<string, object>(),
                ImpressionNumber = 1,
                SessionImpressionNumber = 1,
                CampaignTypeImpressionNumber = 1,
                CampaignTypeSessionImpressionNumber = 1,
                EventNumber = 1,
                SessionEventNumber = 1,
            };

            // Act
            analyticsTracker.TrackImpressionFailEvent(impression, "test");
            await TaskHelper.WaitUntil(() => _interceptedResponseMessage.Result != UnityWebRequest.Result.InProgress, _timeoutCts.Token);

            // Assert
            FailIfTimeoutTokenWasCanceled();
            networkClient
                .Received(2)
                .SendAsync(Arg.Any<string>(), Arg.Any<WebRequestMessage>(), Arg.Any<CancellationToken>()).Forget();
            _interceptedResponseMessage.ResponseCode.Should().Be(200);
            _interceptedResponseMessage.Result.Should().Be(UnityWebRequest.Result.Success);
        }

        [Test]
        [UnitTestTarget(typeof(AnalyticsTracker), nameof(AnalyticsTracker.TrackAppLaunch))]
        public async Task TrackAppLaunch_ThenShouldSucceedSendEventToServer()
        {
            // Arrange
            using var subsystems = CreateAnalyticsTracker(out var analyticsTracker, out var networkClient, out _);
            subsystems.ForEach<IInitializable>(s => s.Initialize());

            // Act
            analyticsTracker.TrackAppLaunch();
            await TaskHelper.WaitUntil(() => _interceptedResponseMessage.Result != UnityWebRequest.Result.InProgress, _timeoutCts.Token);

            // Assert
            FailIfTimeoutTokenWasCanceled();
            networkClient
                .Received(2)
                .SendAsync(Arg.Any<string>(), Arg.Any<WebRequestMessage>(), Arg.Any<CancellationToken>()).Forget();
            _interceptedResponseMessage.ResponseCode.Should().Be(200);
            _interceptedResponseMessage.Result.Should().Be(UnityWebRequest.Result.Success);
        }

        [Test]
        [UnitTestTarget(typeof(AnalyticsTracker), nameof(AnalyticsTracker.TrackAppBackgroundEvent))]
        public async Task TrackSessionEndEvent_ThenShouldSucceedSendEventToServer()
        {
            // Arrange
            using var subsystems = CreateAnalyticsTracker(out var analyticsTracker, out var networkClient, out var prefs);
            subsystems.ForEach<IInitializable>(s => s.Initialize());

            prefs.SessionId.Value = Guid.NewGuid().ToString();

            // Act
            analyticsTracker.TrackAppBackgroundEvent(DateTimeOffset.UtcNow.ToUnixTimeMilliseconds() - 3000, DateTimeOffset.UtcNow.ToUnixTimeMilliseconds());
            await TaskHelper.WaitUntil(() => _interceptedResponseMessage.Result != UnityWebRequest.Result.InProgress, _timeoutCts.Token);

            // Assert
            FailIfTimeoutTokenWasCanceled();
            networkClient
                .Received(2)
                .SendAsync(Arg.Any<string>(), Arg.Any<WebRequestMessage>(), Arg.Any<CancellationToken>()).Forget();
            _interceptedResponseMessage.ResponseCode.Should().Be(200);
            _interceptedResponseMessage.Result.Should().Be(UnityWebRequest.Result.Success);
        }

        [Test]
        public void WhenAnalyticsTrackerStarted_ThanAppLaunchEventSent()
        {
            // Arrange
            using var subsystems = CreateAnalyticsTracker(out var analyticsTracker, out _, out var prefs);

            var platform = new PlatformDefault().AddTo(subsystems);
            new AnalyticsSessionTracker(analyticsTracker, prefs, platform).AddTo(subsystems);
            new SessionCounter(prefs, platform).AddTo(subsystems);

            var appLaunchEvents = 0;
            using var eventObserver = analyticsTracker
                .OnEventSent
                .Where(e => e is EventType.AppLaunch)
                .Subscribe(_ => appLaunchEvents++);

            // Act
            subsystems.ForEach<IInitializable>(s => s.Initialize());

            // Assert
            appLaunchEvents.Should().Be(1);
        }

        [Test]
        public void WhenNewSessionStartedAfterForeground_ThanAppLaunchEventSent()
        {
            // Arrange
            using var subsystems = CreateAnalyticsTracker(out var analyticsTracker, out _, out var prefs);

            var platform = new PlatformDefault().AddTo(subsystems);
            new AnalyticsSessionTracker(analyticsTracker, prefs, platform).AddTo(subsystems);
            var sessionCounter = new SessionCounter(prefs, platform).AddTo(subsystems);
            sessionCounter.Interval.Value = TimeSpan.Zero;

            var appLaunchEvents = 0;
            using var eventObserver = analyticsTracker
                .OnEventSent
                .Where(e => e is EventType.AppLaunch)
                .Subscribe(_ => appLaunchEvents++);

            // Act
            subsystems.ForEach<IInitializable>(s => s.Initialize());
            subsystems.ForEach<IBackgroundListener>(s => s.OnBackground());
            subsystems.ForEach<IForegroundListener>(s => s.OnForeground());

            // Assert
            appLaunchEvents.Should().Be(2);
        }

        [Test]
        public void WhenAppQuitedWithoutGoingToBackground_AndStartedAgain_ThanAppBackgroundEventSent()
        {
            // Arrange
            using (var subsystems = create(out var analyticsTracker, out var platform))
            {
                subsystems.ForEach<IInitializable>(i => i.Initialize());
            }
            var sessionEndEvents = 0;

            // Act
            using (var subsystems = create(out var analyticsTracker, out var platform))
            {
                using var eventObserver = analyticsTracker
                    .OnEventSent
                    .Where(e => e is EventType.AppBackgroundEvent)
                    .Subscribe(_ => sessionEndEvents++);

                subsystems.ForEach<IInitializable>(i => i.Initialize());
            }

            // Assert
            sessionEndEvents.Should().Be(1);

            SubsystemsCollection create(out AnalyticsTracker analyticsTracker, out PlatformDefault platform)
            {
                var subsystems = CreateAnalyticsTracker(out analyticsTracker, out var _, out var prefs);
                platform = new PlatformDefault().AddTo(subsystems);
                new SessionCounter(prefs, platform).AddTo(subsystems);
                new AnalyticsSessionTracker(analyticsTracker, prefs, platform).AddTo(subsystems);
                return subsystems;
            }
        }

        private INetworkClient CreatenetworkClient()
        {
            var networkClient = Substitute.For<INetworkClient>();

            networkClient
                .SendAsync(Arg.Any<string>(), Arg.Any<WebRequestMessage>(), Arg.Any<CancellationToken>())
                .Returns<UniTask<WebResponseMessage>>(
                    async info => await _realNetworkClient.SendAsync(info.ArgAt<string>(0), info.ArgAt<WebRequestMessage>(1), info.ArgAt<CancellationToken>(2)),
                    async info => _interceptedResponseMessage = await _realNetworkClient.SendAsync(info.ArgAt<string>(0), info.ArgAt<WebRequestMessage>(1), info.ArgAt<CancellationToken>(2)));
            return networkClient;
        }

        private GeneralPrefs CreateGeneralPrefs()
        {
            return GeneralPrefs.Create(EditorModeTestsEnvironment.GeneralPrefsPath, _platformAPI);
        }

        private void FailIfTimeoutTokenWasCanceled()
        {
            if (_timeoutCts.IsCancellationRequested)
            {
                Assert.Fail("Timeout token was canceled");
            }
        }

        private SubsystemsCollection CreateAnalyticsTracker(out AnalyticsTracker analyticsTracker, out INetworkClient networkClient, out GeneralPrefs prefs)
        {
            var subsystems = new SubsystemsCollection();
            prefs = CreateGeneralPrefs().AddTo(subsystems);
            networkClient = CreatenetworkClient().AddTo(subsystems);
            var mockServerApi = Substitute.For<IServerApi>();
            analyticsTracker = new AnalyticsTracker(mockServerApi, prefs, EditorModeTestsEnvironment.RootStoragePath).AddTo(subsystems);
            return subsystems;
        }
    }
}